import { createTestingPinia, TestingPinia } from '@pinia/testing'
import { mount } from '@vue/test-utils'
import { defineStore } from 'pinia'
import { describe, expect, test, vi } from 'vitest'
import { ref, h, nextTick } from 'vue'

import { SteeringUser, useHasUserPermission } from '@/stores/auth'
import APermissionGuard from './APermissionGuard.vue'

function createGuard(
  content:
    | string
    | ((slotProps: {
        disabled?: boolean
        areAllChildGuardsHidden: boolean
      }) => ReturnType<typeof h>),
  props: { showPermissions?: string[]; editPermissions?: string[] },
  pinia: TestingPinia,
) {
  return mount(APermissionGuard, {
    global: {
      plugins: [pinia],
    },
    props,
    slots: {
      default: content,
    },
  })
}

function createFakeUser(permissions: string[], isPrivileged = false): SteeringUser {
  return {
    id: 1,
    username: 'her.bert',
    email: 'her.bert@example.org',
    isPrivileged,
    permissions,
    firstName: 'Her',
    lastName: 'Bert',
    isActive: true,
    isStaff: false,
    isSuperuser: false,
    password: '!',
    hostIds: [],
    profile: {
      createdAt: '2024-02-14T15:30:00+01:00',
      createdBy: 'admin',
      updatedAt: '2024-02-14T15:30:00+01:00',
      updatedBy: 'admin',
    },
  }
}

function createPiniaWithFakeUserAuthStore(permissions: string[] = [], isPrivileged = false) {
  const pinia = createTestingPinia({
    createSpy: vi.fn,
  })

  const useAuthStore = defineStore('auth', () => {
    const steeringUser = ref<SteeringUser>(createFakeUser(permissions, isPrivileged))
    return { steeringUser }
  })

  useAuthStore(pinia)

  return pinia
}

// TODO move to different locations once we have dedicated store tests
describe('useHasUserPermission', () => {
  test('Returns true if user has permission.', () => {
    createPiniaWithFakeUserAuthStore(['my-permission'])
    expect(useHasUserPermission(['my-permission']).value).toBeTruthy()
  })

  test('Returns false if the user does not have permission.', () => {
    createPiniaWithFakeUserAuthStore()
    expect(useHasUserPermission(['my-permission']).value).toBeFalsy()
  })

  test('Always returns true if the user is privileged.', () => {
    createPiniaWithFakeUserAuthStore([], true)
    expect(useHasUserPermission(['my-permission']).value).toBeTruthy()
  })
})

describe('APermissionGuard', () => {
  test('Passes disabled attribute if edit permissions are not granted.', () => {
    const pinia = createPiniaWithFakeUserAuthStore()
    const guard = createGuard(
      'Is disabled: {{ params.disabled ? "Yes" : "No" }}',
      { editPermissions: ['my-edit-permission'] },
      pinia,
    )
    expect(guard.text()).toEqual('Is disabled: Yes')
  })

  test('Does not pass a disabled attribute if edit permissions are granted.', () => {
    const pinia = createPiniaWithFakeUserAuthStore(['my-edit-permission'])
    const guard = createGuard(
      'Is disabled: {{ "disabled" in params || params.disabled ? "Yes" : "No" }}',
      { editPermissions: ['my-edit-permission'] },
      pinia,
    )
    expect(guard.text()).toEqual('Is disabled: No')
  })

  test('Does not render content if show permissions are not granted.', () => {
    const pinia = createPiniaWithFakeUserAuthStore()
    const guard = createGuard('my content', { showPermissions: ['my-show-permission'] }, pinia)
    expect(guard.text()).toEqual('')
  })

  test('Does render content if show permissions are granted.', () => {
    const pinia = createPiniaWithFakeUserAuthStore(['my-show-permission'])
    const guard = createGuard('my content', { showPermissions: ['my-show-permission'] }, pinia)
    expect(guard.text()).toEqual('my content')
  })

  test('Passes truthy areAllChildGuardsHidden if every child APermissionGuard component is hidden.', async () => {
    const pinia = createPiniaWithFakeUserAuthStore(['my-show-permission'])
    const guard = createGuard(
      ({ areAllChildGuardsHidden }) => {
        return h('div', [
          h(APermissionGuard, { showPermissions: 'my-other-permission' }, () => ['test: ']),
          areAllChildGuardsHidden ? 'has-no-visible-guards' : 'has-visible-guards',
        ])
      },
      {},
      pinia,
    )
    await nextTick()
    expect(guard.text()).toEqual('has-no-visible-guards')
  })

  test('Passes falsy areAllChildGuardsHidden if any child APermissionGuard component is visible.', async () => {
    const pinia = createPiniaWithFakeUserAuthStore(['my-show-permission'])
    const guard = createGuard(
      ({ areAllChildGuardsHidden }) => {
        return h('div', [
          h(APermissionGuard, { showPermissions: 'my-show-permission' }, () => ['test: ']),
          areAllChildGuardsHidden ? 'has-no-visible-guards' : 'has-visible-guards',
        ])
      },
      {},
      pinia,
    )
    await nextTick()
    expect(guard.text()).toEqual('test: has-visible-guards')
  })
})
